local M = {}
M.type = "auxiliary"
M.relevantDevice = nil
M.defaultOrder = 1000

local constants = {
    -- Conversion rates
    mpsToMph = 2.23694,
    hpToKw = 0.7457,
    torqueToHpCoef = 7120.54016,
    kWhToJ = 3600000,
    sToH = 1.0 / 3600,
    avToRPM = 9.549296596425384,
    rpmToAV = 1.0 / 9.549296596425384,
}

-- lib requires
local mathutil = require("mathutil")
local latch = require("latch")
local pid = require("pid")
local debugutil = require("debugutil")("hybrid", true)
local engineutil = require("engineUtil")
local simpleCurve = require("simplecurve")

local min = math.min
local max = math.max
local abs = math.abs
local maprange = mathutil.maprange
local clamp = mathutil.clamp
local blend = mathutil.blend
local round = mathutil.round
local sign = mathutil.sign
local fnum = debugutil.formatNumber

local engine = nil
local gearbox = nil
local battery = nil
local clutch = nil

local sounds = {
    mg1Whine = nil,
    mg2Whine = nil,
    motorRamp = nil
}

-- Calculated/initial values
local initialWheelBrakeSplit = {}
local torqueData = {
    maxElectricTorque = 0,
    fullBatteryElectricTorque = 0,
    maxEngineTorque = 0,
    maxCombinedTorque = 0,
    engineIdleTorque = 0,
    engineMaxTorqueRpm = 0,
    engineTorqueCurve = {},
}
local engineTorqueData

-- Quick access upvalues
local engineOn = false
local watertemp = 0
local speedMph = 0
local batteryLevel = 1
local totalPowerOutKw = 0
local throttle = 0
local throttleAggressiveness = 0
local gear = "P"
local brakeInput = 0

-- Other state variables/upvalues
local triggerEngineDemandReason = "none"
local engineDemandReason = "none"
local lastEngineDemandedOn = false
local miscEngineDemandReason = "none"
local miscEngineDemand = false
local waterTempEngineDemand = false
local state = "init"
local nextState = nil
local lastState = nil
local timeSinceEngineStart = 5
local timeSinceEngineCanShutOff = 0
local timeSinceAppliedBrakes = 0
local ecvtDriveMode = 0
local highSpeedCoast = false
local startingTime = 0
local isInChargeup = false
local lastThrottle = 0
local chargeThrottle = 0
local disableRegen = false
local timeSinceAbsHit = 0
local timeSinceInit = 0
local mg2TorqueAdjustment = 0
local engineDeload = 0
local desiredRegen = 0
local availableMG1Power = 0
local availableMG2Power = 0
local targetEngineRpm = 0
local availablePriorityEvPower = 0
local availableSecondaryEvPower = 0
local maxBatteryPower = 0
local availableBatteryPower = 0
local batteryWorkingRatio = 0
local requestedChargePower = 0
    
local function getProportionalBatteryRatio(minLevelBias, maxLevelBias)
    minLevelBias = minLevelBias or 0
    maxLevelBias = maxLevelBias or 0

    local minPowerLevelThreshold = blend(M.minBatteryPowerLevel, M.maxBatteryPowerLevel, minLevelBias)
    local maxPowerLevelThreshold = blend(M.minBatteryPowerLevel, M.maxBatteryPowerLevel, 1 - maxLevelBias)

    return maprange(batteryLevel, minPowerLevelThreshold, maxPowerLevelThreshold)
end

local function calculateElectricLimits(dt)
    -- Calculate the coefficient of battery charge left (more power from the battery
    -- can be used at low state-of-charge in EV mode)
    if state == "evmode" then
        batteryWorkingRatio = maprange(batteryLevel, 0.15, M.maxBatteryPowerLevel)
    else
        batteryWorkingRatio = getProportionalBatteryRatio()
    end
    
    -- Calculate battery power limits
    local availableBatteryPowerCoef
    
    if state == "startup" then
        availableBatteryPowerCoef = maprange(batteryLevel, 0, 0.05)
    else
        availableBatteryPowerCoef = maprange(batteryWorkingRatio, 0, 0.05)
    end
    
    maxBatteryPower = ((battery.jbeamData.maxPowerOutput or 1e5) - M.idlePowerDraw)
    availableBatteryPower = availableBatteryPowerCoef * maxBatteryPower
    
    -- Figure out some general values about what the electric subsystem can currently provide
    if state == "evmode" then
        availablePriorityEvPower = availableBatteryPower
        availableSecondaryEvPower = 0
    else
        local evPowerAvailableFromGenerator = M.mg1PowerGeneratedSmoother:get(max(0, -gearbox.motorGenerator1Power), dt)
        -- TODOFIX
        local priorityBatteryPowerCoef = M.maxEngineOnPriorityRatio * maprange(batteryWorkingRatio, M.targetRelativeBatteryLevel, M.targetRelativeBatteryLevel + M.targetBatteryLevelTolerance)
        
        availablePriorityEvPower = availableBatteryPower * priorityBatteryPowerCoef
        availableSecondaryEvPower = evPowerAvailableFromGenerator + availableBatteryPower
    end
end

local function calculateTorqueStatistics()
    -- Calculate some torque performance statistics
        
    -- Raw max torques from gearbox and engine
    local maxFinalMG2Torque = gearbox:getMaxMG2Torque(availablePriorityEvPower + availableSecondaryEvPower) * gearbox.finalMG2GearRatio
    
    if state == "evmode" then
        local maxFinalMG1Torque = gearbox:getRingTorqueFromSunTorque(gearbox:getMaxMG1Torque()) * gearbox.finalRingGearRatio
        
        torqueData.maxElectricTorque = maxFinalMG2Torque + maxFinalMG1Torque
        torqueData.fullBatteryElectricTorque = torqueData.maxElectricTorque
        torqueData.maxEngineTorque = 0
        torqueData.maxCombinedTorque = torqueData.maxElectricTorque
    else
        -- Figure out compound torques
        local maxEngineTorque = engine.torqueData.maxTorque
        local maxSunGearTorque = gearbox:getMaxMG1Torque(true)
        local maxEngineTorqueToCarrier = maxEngineTorque * gearbox.inputGearRatio
        local maxEngineTorqueToSun = abs(gearbox:getSunTorqueFromCarrierTorque(maxEngineTorqueToCarrier))
        local maxUsableEngineTorqueToSun = min(maxEngineTorqueToSun, maxSunGearTorque)
        local maxEngineTorqueToOutput = abs(gearbox:getRingTorqueFromSunTorque(maxUsableEngineTorqueToSun)) * gearbox.finalRingGearRatio
        
        torqueData.maxElectricTorque = maxFinalMG2Torque
        torqueData.fullBatteryElectricTorque = gearbox:getMaxMG2Torque(maxBatteryPower) * gearbox.finalMG2GearRatio
        torqueData.maxEngineTorque = maxEngineTorqueToOutput
        torqueData.maxCombinedTorque = torqueData.maxEngineTorque + torqueData.maxElectricTorque
    end
    
    -- Figure out engine torque characteristics
    local curve = engine.torqueData.curves[engine.torqueData.finalCurveName]
    local idleTorque = curve.torque[engine.idleRPM]
    
    torqueData.engineIdleTorque = idleTorque
    torqueData.engineMaxTorqueRpm = engine.torqueData.maxTorqueRPM
    torqueData.engineTorqueCurve = curve.torque
end

local function getTorqueData()
    return torqueData
end

local function printTorqueData()
    local curTorqueData = getTorqueData()
    
    debugutil.log("I", "Max Electric Torque: " .. fnum(curTorqueData.maxElectricTorque))
    debugutil.log("I", "Max Full-Battery Electric Torque: " .. fnum(curTorqueData.fullBatteryElectricTorque))
    debugutil.log("I", "Max Usable Engine Torque: " .. fnum(curTorqueData.maxEngineTorque))
    debugutil.log("I", "Max Optimal Total Torque: " .. fnum(curTorqueData.maxCombinedTorque))
    debugutil.log("I", "Max Engine Idle Torque: " .. fnum(curTorqueData.engineIdleTorque))
    debugutil.log("I", "Engine Max Torque RPM: " .. fnum(curTorqueData.engineMaxTorqueRpm))
end

local function requestEngine(reason)
    if miscEngineDemand then return end
    
    miscEngineDemandReason = reason
    miscEngineDemand = true
end

local function isEngineDemandedOn()
    local inputThrottle = electrics.values.throttleOverride or input.throttle
    local minRunTime
    
    if watertemp < 57 then
        minRunTime = M.engineMinRunTime * 2
    else
        minRunTime = M.engineMinRunTime
    end

    if state == "chargemode" then
        engineDemandReason = "charge mode"
        return true
    end
    if not M.enableAutoStop then
        engineDemandReason = "auto stop disabled"
        return true
    end
    if watertemp < 30 or waterTempEngineDemand then
        engineDemandReason = "temp"
        waterTempEngineDemand = true
        return true
    else
        
    end -- engine must be warmed up
    
    if gear == "R" and batteryLevel >= blend(M.minBatteryPowerLevel, M.maxBatteryPowerLevel, 0.2) and timeSinceEngineStart >= minRunTime then
        -- After this point, if in R and battery >= 0.4, force engine off
        engineDemandReason = "none (forced off due to reverse)"
        return false
    end
    
    if speedMph > M.maxEvOnlySpeed then
        engineDemandReason = "speed"
        return true
    end -- too fast
    if obj:getDirectionVector().z > M.maxEvOnlyPitch then
        engineDemandReason = "vehicle pitch"
        return true
    end -- pitched too high up (hill start)
    if batteryLevel < M.minBatteryPowerLevel or isInChargeup then
        engineDemandReason = "battery level"
        isInChargeup = true
        return true
    end -- battery too low
    if throttleAggressiveness > M.maxEvOnlyThrottleAggressiveness then
        engineDemandReason = "throttle"
        return true
    end
    if miscEngineDemand then
        engineDemandReason = miscEngineDemandReason
        return true
    end
    
    local actualRequestedBrake = max(brakeInput, desiredRegen)
    
    if engineOn and (
        not (electrics.values.parkingbrake == 1 or actualRequestedBrake >= M.autoStopBrakeThreshold) or
        timeSinceAppliedBrakes < M.autoStopBrakeTime
    ) then
        -- this check only applies when not in P
        if gear ~= "P" then
            engineDemandReason = "brakes"
            return true
        end
    end

    -- Easy checks are done; dynamic checks now

    -- This is the percentage of the max EV-only kW output that's available to us,
    -- based on battery level

    if engineOn then
        if inputThrottle > M.engineShutOffThrottleThreshold then
            engineDemandReason = "accelerating with engine on"
            return true
        end -- once engine is on, must coast again to shut off
    else
        if gear ~= "R" and gearbox.motorGenerator2Power > M.maxEvOnlyPriorityPower then
            engineDemandReason = "priority power exceeded"
            return true
        end
    end

    -- Put this last so it doesn't override other reasons
    if engineOn and timeSinceEngineStart < minRunTime then
        engineDemandReason = "time since start"
        return true
    end -- engine has to run for at least 4 seconds

    engineDemandReason = "none"
    return false -- by default, the engine is off
end

local function updateValues(dt)
    if not gearbox or not engine or not battery then
        return
    end
    
    local throttleAggressivenessInst = (input.throttle - lastThrottle) / dt
    
    miscEngineDemandReason = "none"
    miscEngineDemand = false
    lastThrottle = input.throttle
    engineOn = not (engine.isStalled or engine.ignitionCoef < 0.01 or electrics.values.rpm < engine.idleRPM / 2)
    watertemp = electrics.values.watertemp
    speedMph = electrics.values.wheelspeed * constants.mpsToMph
    batteryLevel = battery.remainingRatio
    gear = electrics.values.gear
    mg2TorqueAdjustment = 0
    brakeInput = input.brake
    
    -- Rolling average
    throttleAggressiveness = throttleAggressiveness + (throttleAggressivenessInst - throttleAggressiveness) / 10
    
    local throttleRate = maprange(
        abs(throttleAggressiveness),
        M.minThrottleAggressivenessRamp,
        M.maxThrottleAggressivenessRamp,
        M.minThrottleAggressivenessFactor,
        M.maxThrottleAggressivenessFactor
    )
    
    debugutil.putValue("Throttle Rate", throttleRate)
    
    throttle = electrics.values.throttleOverride or M.throttleSmoother:get(input.throttle, dt * throttleRate)

    if engineOn then
        timeSinceEngineStart = timeSinceEngineStart + dt
        
        if watertemp >= 57 then
            waterTempEngineDemand = false
        end
    else
        timeSinceEngineStart = 0
    end

    local actualRequestedBrake = max(brakeInput, desiredRegen)
    
    if M.autoStopBrakeThreshold > 0 and (actualRequestedBrake >= M.autoStopBrakeThreshold or electrics.values.parkingbrake == 1) and speedMph <= M.maxEvOnlySpeed then
        timeSinceAppliedBrakes = timeSinceAppliedBrakes + dt
    else
        timeSinceAppliedBrakes = 0
    end

    if isInChargeup and batteryLevel >= blend(M.minBatteryPowerLevel, M.maxBatteryPowerLevel, 0.2) then
        isInChargeup = false
    end
end

local function updateState(dt)
    local engineDemandedOn = isEngineDemandedOn()

    if engineDemandedOn and not lastEngineDemandedOn then
        triggerEngineDemandReason = engineDemandReason
    elseif not engineDemandedOn then
        triggerEngineDemandReason = "none"
    end
    
    lastEngineDemandedOn = engineDemandedOn
    
    debugutil.putValue("Coolant Temp.", electrics.values.watertemp)
    debugutil.putValue("Engine demanded on", tostring(engineDemandedOn))
    debugutil.putValue("Engine on reason", engineDemandReason)
    debugutil.putValue("Engine triggered on by", triggerEngineDemandReason)
    debugutil.putValue("State", state)

    -- keep things like daytime-running-lights on
    electrics.values.running = 1

    -- prevent engine starter from working (since the eCVT starts it)
    engine:deactivateStarter()
    engine.starterDisabled = true

    if state == "init" then
        timeSinceInit = timeSinceInit + dt

        if timeSinceInit >= 1 then
            if engineOn then
                nextState = "running"
            else
                nextState = "stopped"
            end
        end
    elseif state == "evmode" then
        engine.ignitionCoef = 0 -- Kill engine
        engine.isDisabled = true -- Prevent the user from manually restarting the engine
        electrics.values.checkengine = false -- Keep this from coming on (usually tied to engine.isDisabled)
        gearbox:setInputDisengaged(true)
        
        if electrics.values.chargeMode == 1 then
            if gear == "P" then
                electrics.values.evMode = 0
            else
                gui.message({txt = "Charge mode can only be enabled in P gear"}, 2, "vehicle.hybrid.chargeMode")
                electrics.values.chargeMode = 0
            end
        end

        if electrics.values.evMode ~= 1 then
            if M.forceEvMode then
                electrics.values.evMode = 1
            else
                gui.message({txt = "EV mode disabled"}, 2, "vehicle.hybrid.evMode")
                gearbox:setInputDisengaged(false)
                nextState = "init"
            end
        end
        
        if batteryWorkingRatio < 0.025 and not M.forceEvMode then
            electrics.values.evMode = 0
            gui.message({txt = "Battery too low; EV mode disabled"}, 2, "vehicle.hybrid.evMode")
            gearbox:setInputDisengaged(false)
            nextState = "init"
        end
    elseif state == "running" then
        if M.maxEngineDeload > 0 then
            engine.ignitionCoef = 1 - M.engineDeloadMaxIgnitionCut * (engineDeload / M.maxEngineDeload)
        else
            engine.ignitionCoef = 1
        end
        
        if not engineOn then
            -- engine died?

            timeSinceInit = 0
            nextState = "init"
        elseif electrics.values.evMode == 1 then
            if batteryWorkingRatio > 0.026 then
                gui.message({txt = "EV mode enabled"}, 2, "vehicle.hybrid.evMode")
                nextState = "evmode"
            else
                gui.message({txt = "Battery too low; EV mode not available"}, 2, "vehicle.hybrid.evMode")
                electrics.values.evMode = 0
            end
        elseif electrics.values.chargeMode == 1 then
            if gear == "P" then
                gui.message({txt = "Charge mode enabled"}, 2, "vehicle.hybrid.chargeMode")
                chargeThrottle = 0
                nextState = "chargemode"
            else
                gui.message({txt = "Charge mode can only be enabled in P gear"}, 2, "vehicle.hybrid.chargeMode")
                electrics.values.chargeMode = 0
            end
        else
            if engineDemandedOn then
                timeSinceEngineCanShutOff = 0
            else
                if gear ~= "N" then -- engine keeps state in N
                    timeSinceEngineCanShutOff = timeSinceEngineCanShutOff + dt

                    if timeSinceEngineCanShutOff > 1 then
                        engine.ignitionCoef = 0 -- Kill engine
                        engine.isDisabled = true -- Prevent the user from manually restarting the engine
                        electrics.values.checkengine = false -- Keep this from coming on (usually tied to engine.isDisabled)

                        nextState = "stopped"
                    end
                end
            end
        end
    elseif state == "chargemode" then
        local inverseRelativePowerLevel = maprange(batteryWorkingRatio, 0.95, 1.0, 1.0, 0.05)
        local maxFastChargePower = M.maxFastChargePower or gearbox.mg1PowerRating
        local curGenOutput = -min(totalPowerOutKw, 0)
        local targGenOutput = inverseRelativePowerLevel * maxFastChargePower

        if lastState ~= state then
            M.chargeThrottlePid:reset()
        end
        
        chargeThrottle = clamp(M.chargeThrottlePid:get(targGenOutput, curGenOutput, dt), 0, 1)

        if gear ~= "P" or electrics.values.evMode == 1 or batteryWorkingRatio >= 0.998 then
            electrics.values.chargeMode = 0
        end

        if electrics.values.chargeMode ~= 1 then
            gui.message({txt = "Charge mode disabled"}, 2, "vehicle.hybrid.chargeMode")
            nextState = "init"
        end
    elseif state == "stopped" then
        timeSinceEngineCanShutOff = 0
        engine.ignitionCoef = 0
        electrics.values.checkengine = false -- Keep this from coming on (usually tied to engine.isDisabled)
        electrics.values.iceThrottle = 0
        
        if electrics.values.chargeMode == 1 then
            if gear == "P" then
                engineDemandedOn = true
            else
                gui.message({txt = "Charge mode can only be enabled in P gear"}, 2, "vehicle.hybrid.chargeMode")
                electrics.values.chargeMode = 0
            end
        end

        if electrics.values.evMode == 1 then
            gui.message({txt = "EV mode enabled"}, 2, "vehicle.hybrid.evMode")
            nextState = "evmode"
        elseif gear ~= "N" and engineDemandedOn then
            engine.isDisabled = false -- allow engine to restart

            if electrics.values.rpm < engine.idleRPM then
                nextState = "startup"
                timeSinceEngineStart = 0
                startingTime = 0
            else
                engine.ignitionCoef = 1 -- engine already spinning, engage ignition
                nextState = "running"
            end
        end
    elseif state == "startup" then
        local targetStarterRpm = M.engineMinIdleRpm
        local actualRPM = electrics.values.rpm
        local rpmError = targetStarterRpm - actualRPM
        local targetStarterTorque = maprange(rpmError, 0, 200, 0, M.engineStarterTorque)
        
        engine.ignitionCoef = maprange(rpmError, 500, 300, 0, 1)
        electrics.values.iceThrottle = 0
        M.engineThrottleSmoother:reset()
        startingTime = startingTime + dt
        
        if engineOn then
            M.driveThrottlePid:setILock(false)
            
            local targetIceThrottle = M.driveThrottlePid:get(targetEngineRpm, actualRPM, dt)
            
            electrics.values.iceThrottle = clamp(targetIceThrottle, 0, 1)
            engine.ignitionCoef = 1
            
            if timeSinceEngineStart > M.engineStartupIdleTime then
                targetStarterTorque = 0
                nextState = "running"
            end
        else
            M.driveThrottlePid:reset()
            
            if startingTime > 3 then
                -- Something's not right; retry at init
                
                targetStarterTorque = 0
                timeSinceInit = 0
                nextState = "init"
            end
        end
        
        local starterTorque = M.engineStarterSmoother:get(targetStarterTorque, dt)
        local starterCoef
        
        gearbox:setMG1PowerOverride(maprange(batteryLevel, 0.05, 0.1, 0, gearbox.mg1PowerRating))
        gearbox:setMG1TorqueDirection(1)
        
        if gearbox.sunGearAV < (-gearbox.generatorMinFullLoadAV * 5) then
            starterCoef = -starterTorque
        else
            starterCoef = starterTorque
        end 
        
        gearbox:setRequestedMG1Torque(starterCoef)
    end
    
    lastState = state
end

local function updateBattery(dt)
    local mg1Kw = gearbox.motorGenerator1Power
    local mg2Kw = gearbox.motorGenerator2Power
    
    -- If negative (generating power), multiply by generator efficiency (power loss from motors to battery)
    -- If positive (consumint power), divide by generator efficiency (power loss from battery to motors)
    if mg1Kw < 0 then
        mg1Kw = mg1Kw * M.regenEfficiency
    else
        mg1Kw = mg1Kw / M.regenEfficiency
    end
    
    if mg2Kw < 0 then
        mg2Kw = mg2Kw * M.regenEfficiency
    else
        mg2Kw = mg2Kw / M.regenEfficiency
    end
    
    totalPowerOutKw = M.idlePowerDraw + mg1Kw + mg2Kw

    local totalPowerOutKj = totalPowerOutKw * dt

    debugutil.putValue("Battery power (kW)", totalPowerOutKw)
    debugutil.putValue("Generator power (kW)", mg1Kw)

    battery.storedEnergy = mathutil.clamp(battery.storedEnergy - totalPowerOutKj * 1000, 0, battery.initialStoredEnergy)
end

local function updateSounds(dt)
    -- Update sounds
    local mg1Power = abs(gearbox.motorGenerator1Power)
    local mg2Power = abs(gearbox.motorGenerator2Power)
    local mg1AV = abs(gearbox.sunGearAV)
    local mg2AV = abs(gearbox.ringGearAV)
    local mg1InverterPitch = maprange(mg1AV, M.inverterPitchMinSpeed, M.inverterPitchMaxSpeed, M.inverterMinPitch, M.inverterMaxPitch)
    local mg1InverterEffort = mg1Power / M.inverterMaxVolumeKw
    local mg1InverterVolume = M.inverterBaseVolume * clamp(mg1InverterEffort, 0.0, 1.0)
    local mg2InverterPitch = maprange(mg2AV, M.inverterPitchMinSpeed, M.inverterPitchMaxSpeed, M.inverterMinPitch, M.inverterMaxPitch)
    local mg2InverterEffort = mg2Power / M.inverterMaxVolumeKw
    local mg2InverterVolume = M.inverterBaseVolume * clamp(mg2InverterEffort, 0.0, 1.0)

    obj:setVolumePitch(sounds.mg1Whine, mg1InverterVolume, M.mg1PitchSmoother:get(mg1InverterPitch, dt))
    obj:setVolumePitch(sounds.mg2Whine, mg2InverterVolume, M.mg2PitchSmoother:get(mg2InverterPitch, dt))

    local motorPitch =
        maprange(
        speedMph,
        M.motorPitchMinSpeed,
        M.motorPitchMaxSpeed,
        M.motorMinPitch,
        M.motorMaxPitch,
        false
    )
    local motorVolume =
        maprange(
        speedMph,
        M.motorVolumeMinSpeed,
        M.motorVolumeMaxSpeed,
        0,
        M.motorBaseVolume,
        false
    )

    obj:setVolumePitch(sounds.motorRamp, motorVolume, motorPitch)
end

local function updateGUI(dt)
    local maxKw
    
    if battery and battery.jbeamData and (battery.jbeamData.maxPowerOutput or battery.jbeamData.maxPowerInput) then
        local maxOutput = battery.jbeamData.maxPowerOutput or 0
        local maxInput = battery.jbeamData.maxPowerInput or 0
        
        maxKw = max(maxOutput, maxInput)
    else
        maxKw = gearbox and max(gearbox.mg1PowerRating, gearbox.mg2PowerRating) or 0
    end
    
    local str = {
        soc = batteryWorkingRatio,
        kw = totalPowerOutKw,
        maxKw = maxKw,
        minSoc = 0.0,
        maxSoc = 1.0,
    }

    gui.send("hybrid", str)
    debugutil.sendStreams()
end

local function updateGFX(dt)
    updateValues(dt)

    local iceThrottle = 0.0
    
    if gearbox and engine and battery and clutch then
        calculateElectricLimits(dt)
        calculateTorqueStatistics()
        
        -- Once we know we have all the required components for running,
        -- we can start to perform our (very complex and abundant) hybrid
        -- calculations
        
        -- Calculate some rudimentary locals
        local actualSpeedSign = mathutil.sign(electrics.values.avgWheelAV)
        local desiredSpeedSign = gear == "R" and -1 or 1
        
        -- Pull in some measurements on the gearbox
        local gearboxMaxMG2Torque = gearbox:getMaxMG2Torque()
        local gearboxMaxMG1TorqueUnchecked = gearbox:getMaxMG1Torque(true)
        local gearboxMaxMG2TorqueUnchecked = gearbox:getMaxMG2Torque(true)
        
        -- If there is a maximum power input on the battery, we can't exceed it while regenerating
        -- Instead, we have to figure out the maximum regen coefficient that we can apply without
        -- exceeding the battery's max power input
        local maxRegenPercentOfMg2PowerRating = 1
        
        if battery.jbeamData.maxPowerInput then
            local gearboxMaxRegenKw = mathutil.motorTorqueToKw(gearboxMaxMG2Torque, abs(gearbox.motorGenerator2AV))
            local batteryPowerInputCoef = maprange(batteryLevel, 0.9, 0.95, 1, 0)
            local batteryPowerInputAvailable = battery.jbeamData.maxPowerInput * batteryPowerInputCoef
            
            -- Calculate what percentage of regen we can use, taking into account
            -- the maximum amount of power the battery can currently accept, minus
            -- however much power the engine is putting into the battery
            maxRegenPercentOfMg2PowerRating = (batteryPowerInputAvailable) / (gearboxMaxRegenKw * M.regenEfficiency)
        end
        
        local isInDriveGear = gear ~= "P" and gear ~= "N"
        local maxRegen = not disableRegen and min(M.maxRegenCoef, maxRegenPercentOfMg2PowerRating) or 0
        
        -- Calculate some throttle input variables
        local throttleOffset = max(maxRegen * (gear == "L" and 0.25 or M.evCoastRegen), 0.0001)
        local posThrottle = math.pow( maprange(throttle, throttleOffset, 1), 1 / M.throttleCurveExponent )
        local negThrottle = maprange(throttle, throttleOffset, 0)
        local requestedThrottle = posThrottle
        local creepThrottle = 0
        local requestedRegen = 0

        -- Fill in some debug values
        debugutil.putValue("Battery SOC % (usable)", math.floor(getProportionalBatteryRatio() * 100))
        debugutil.putValue("Throttle % EV thresh.", math.floor(throttle * 100 / M.maxEvOnlyThrottle))
        
        if brakeInput < 0.25 and electrics.values.parkingbrake < 1 and gear ~= "P" and gear ~= "N" then
            -- Apply creep throttle
            
            creepThrottle = maprange(speedMph, 0, 5, M.evCreepThrottle, 0)
        end
        
        -- Apply hill assist
        local hillAssistAngleCoef = maprange(obj:getDirectionVector().z * desiredSpeedSign, M.minHillAssistPitch, M.maxHillAssistPitch)
        local hillAssistSpeedCoef = maprange(speedMph * actualSpeedSign * desiredSpeedSign, 1e-10, 0.05, 1, 0)
        local hillAssistTorqueCoef = maprange(abs(gearbox.outputTorque1), 0, torqueData.maxCombinedTorque, 1, 0)
        local hillAssistBrakeHold = hillAssistAngleCoef * hillAssistSpeedCoef * hillAssistTorqueCoef * M.hillAssistBrakeHoldAmount
        
        input.brake = max(brakeInput, hillAssistBrakeHold)
        electrics.values.brake = max(electrics.values.brake, hillAssistBrakeHold)

        if input.brake > 0 then
            -- Updating while brake pedal is applied
            
            if not isInDriveGear or gear == "R" then
                -- No regen in P, R, or N
                
                for i = 0, wheels.wheelCount - 1, 1 do
                    local wheel = wheels.wheels[i]

                    if wheel then
                        -- Reset brake split coef
                        wheel.brakeInputSplit = initialWheelBrakeSplit[i][1]
                        wheel.brakeSplitCoef = initialWheelBrakeSplit[i][2]
                    end
                end

                debugutil.putValue("Friction % of full brakes", math.floor(input.brake * 100))
            else
                -- Blend friction/regen brakes in any other gear
                
                local maxBrakeTorque = 0.0
                local absActive = false

                for i = 0, wheels.wheelCount - 1, 1 do
                    local wheel = wheels.wheels[i]

                    if wheel then
                        -- Find total braking torque
                        if wheel.brakeTorque > maxBrakeTorque then
                            maxBrakeTorque = wheel.brakeTorque
                        end

                        -- Determine if ABS active
                        if wheel.absActive then
                            absActive = true
                        end
                    end
                end

                if disableRegen then
                    if not absActive then
                        timeSinceAbsHit = timeSinceAbsHit + dt

                        if timeSinceAbsHit > 1.0 then
                            disableRegen = false
                            timeSinceAbsHit = 0
                        end
                    end
                else
                    if absActive then
                        timeSinceAbsHit = timeSinceAbsHit + dt

                        if timeSinceAbsHit > 0.5 then
                            disableRegen = true
                            timeSinceAbsHit = 0
                        end
                    else
                        timeSinceAbsHit = 0
                    end
                end

                -- Combine regen input from brake pedal and throttle pedal (one-pedal driving)
                local regenBrakePedal = maprange(brakeInput, 0, M.regenPercentOfBrakingForce)
                local regenAccelPedal = gear == "L" and negThrottle or 0
                
                desiredRegen = clamp(regenBrakePedal + regenAccelPedal, 0, 1) * maprange(speedMph, 0, M.minRegenFadeSpeed)
                requestedRegen = maxRegen * desiredRegen

                -- Calculate amount of friction brakes needed to cover brake demand
                local actualRegenCoef = requestedRegen < 0.01 and 0 or clamp((gearbox.motorGenerator2Torque * actualSpeedSign) / (-requestedRegen * gearboxMaxMG2Torque), 0, 1)
                local actualEffectiveRegen = actualRegenCoef * maxRegen * M.regenPercentOfBrakingForce
                local brakesCoveredByRegen = maprange(input.brake + (regenAccelPedal * M.regenPercentOfBrakingForce), actualEffectiveRegen, 1, 1, 0)
                local neededFrictionBrake = maprange(input.brake + (regenAccelPedal * M.regenPercentOfBrakingForce), brakesCoveredByRegen, 1)

                debugutil.putValue("Max Regen", maxRegen)
                debugutil.putValue("Regen % of full brakes", math.floor(actualEffectiveRegen * 100))
                debugutil.putValue("Regen % of full regen", math.floor(actualRegenCoef * 100))
                debugutil.putValue("Friction % of full brakes", math.floor(neededFrictionBrake * 100))

                -- Apply more friction brakes on wheels without regen to balance them
                for i = 0, wheels.wheelCount - 1, 1 do
                    local wd = wheels.wheels[i]

                    wd.brakeInputSplit = 1 - brakesCoveredByRegen

                    if wd.isPropulsed then
                        -- Propulsed wheels are affected by regenerative braking

                        wd.brakeSplitCoef = initialWheelBrakeSplit[i][2] * max(neededFrictionBrake - actualRegenCoef, 0)
                    else
                        -- Unpropulsed wheels are not

                        wd.brakeSplitCoef = initialWheelBrakeSplit[i][2] * neededFrictionBrake
                    end
                end
            end
        else
            disableRegen = false

            if isInDriveGear then
                -- Regenerative braking when letting off accelerator (tiny bit in D, full in L)
                local zeroThrottleRegen = gear == "L" and 1 or M.evCoastRegen
                
                desiredRegen = negThrottle * zeroThrottleRegen * maprange(speedMph, 0, M.minRegenFadeSpeed)
                requestedRegen = maxRegen * desiredRegen
            else
                desiredRegen = 0
                requestedRegen = 0
            end
        end
        
        local mg1Dir = 1
        local mg1Torque = 0.0
        local mg2Torque = 0.0
        
        if state == "init" then
            iceThrottle = 0
        elseif gear == "N" then
            iceThrottle = throttle
        elseif state == "chargemode" then
            iceThrottle = chargeThrottle
            
            local chargeTargetRpm = M.fastChargeRpm or M.engineMaxRpm
            local desiredGenPowerCoef = maprange(electrics.values.rpm, chargeTargetRpm - M.engineLoadRpmTolerance, chargeTargetRpm)
            
            mg1Torque = -desiredGenPowerCoef * gearboxMaxMG1TorqueUnchecked
        else
            -- TOO MANY CALCULATIONS, SEND HELP
            
            -- Throttle-related variables
            local normalizedThrottle = requestedThrottle
            local requestedTorque = 0
            local requestedElectricOutputTorque = 0
            local requestedEngineOutputTorque = 0
            local targetEngineRpmCoef = 0
            local engineLoad = 0
            
            -- Prevent gearbox or vehicle from going overspeed
            local mg1OverspeedCoef = 0
            local mg2OverspeedCoef = 0
            local vehicleOverspeedCoef = 0
            local maxVehicleSpeed = gear == "R" and M.maxVehicleSpeedReverse or M.maxVehicleSpeed
            local maxVehicleSpeedFalloff = gear == "R" and M.maxVehicleSpeedFalloffReverse or M.maxVehicleSpeedFalloff
            
            -- Limit vehicle speed to spec (with falloff)
            if type(maxVehicleSpeed) == "number" then
                vehicleOverspeedCoef = maprange(speedMph, maxVehicleSpeed, maxVehicleSpeed + maxVehicleSpeedFalloff)
            end
            
            -- Prevent generator motor (MG1) from negative overspeeding ONLY in evmode (engine RPM will prevent this when it's on)
            if state == "evmode" and gearbox.motorGenerator1AV <= 0 then
                mg1OverspeedCoef = maprange(gearbox.motorGenerator1AV, gearbox.minMG1AV + 100, gearbox.minMG1AV)
            end
            
            -- Prevent drive motor (MG2) from overspeeding in either direction
            if gearbox.motorGenerator2AV >= 0 then
                mg2OverspeedCoef = maprange(gearbox.motorGenerator2AV, gearbox.maxMG2AV - 100, gearbox.maxMG2AV)
            else
                mg2OverspeedCoef = maprange(gearbox.motorGenerator2AV, gearbox.minMG2AV + 100, gearbox.minMG2AV)
            end
            
            local overspeedCoef = M.throttleLimiterSmoother:get(mathutil.maxall(
                mg1OverspeedCoef,
                mg2OverspeedCoef,
                vehicleOverspeedCoef
            ), dt)
            
            normalizedThrottle = min(normalizedThrottle, 1 - overspeedCoef)
            
            -- Calculate RPM range for engine based on required sun gear speeds
            local minRpm = max(M.engineMinIdleRpm, (gearbox:getMinCarrierAV() * gearbox.inputGearRatio) * constants.avToRPM)
            local maxRpm = min(M.engineMaxRpm, (gearbox:getMaxCarrierAV() * gearbox.inputGearRatio) * constants.avToRPM)
            local actualRpm = M.engineActualRpmSmoother:get(round(electrics.values.rpm), dt)
            
            -- Figure out max possible engine torque at its current RPM, as well as the torque it's currently
            -- applying on\the different transmission components
            local achievableEngineTorque = engineutil.getInstantMaxEngineTorque(engine)
            local actualEngineTorque = abs(engine.outputTorque1)
            local actualEngineTorqueOnCarrier = actualEngineTorque * gearbox.inputGearRatio
            local actualEngineTorqueOnRing = abs(gearbox:getRingTorqueFromCarrierTorque(actualEngineTorqueOnCarrier))
            local actualEngineTorqueOnSun = abs(gearbox:getSunTorqueFromCarrierTorque(actualEngineTorqueOnCarrier))
            local actualEngineTorqueOnOutput = actualEngineTorqueOnRing * gearbox.finalRingGearRatio
            
            maxRpm = max(minRpm, maxRpm)
            
            if isInDriveGear then
                -- Update while in a driving gear (anything other than Park or Neutral)
                
                -- Map throttle to total capable torque output
                if type(M.evOnlyThrottleDetent) == "number" then
                    -- There's a specific detent in the accelerator where the engine will kick in, if the battery is full
                    -- This makes it easier to keep the vehicle in EV mode with a full battery
                    
                    local requestedEvTorque = maprange(normalizedThrottle, 0, M.evOnlyThrottleDetent) * torqueData.fullBatteryElectricTorque
                    local requestedEngineTorque = maprange(normalizedThrottle, M.evOnlyThrottleDetent, 1) * torqueData.maxEngineTorque
                    
                    requestedTorque = requestedEvTorque + requestedEngineTorque
                else
                    requestedTorque = normalizedThrottle * torqueData.maxCombinedTorque
                end
                
                -- Calculate available EV torque and distribute torque demand between motors and engine
                local availablePriorityEvOutputTorque
                local availableSecondaryEvOutputTorque
                
                if state == "evmode" then
                    local gearboxMaxMG1Torque = gearbox:getMaxMG1Torque(availablePriorityEvPower)
                    local availableMG1OutputTorque = gearbox:getRingTorqueFromSunTorque(gearboxMaxMG1Torque) * gearbox.finalRingGearRatio
                    local availableMG2OutputTorque = gearbox:getMaxMG2Torque(availablePriorityEvPower) * gearbox.finalMG2GearRatio
                    
                    availablePriorityEvOutputTorque = availableMG1OutputTorque + availableMG2OutputTorque
                    availableSecondaryEvOutputTorque = 0
                else
                    availablePriorityEvOutputTorque = gearbox:getMaxMG2Torque(availablePriorityEvPower) * gearbox.finalMG2GearRatio
                    availableSecondaryEvOutputTorque = gearbox:getMaxMG2Torque(availableSecondaryEvPower) * gearbox.finalMG2GearRatio
                end
                
                debugutil.putValue("availablePriorityEvOutputTorque", availablePriorityEvOutputTorque)
                
                local torqueStillNeeded = requestedTorque
                local creepTorque = creepThrottle * torqueData.maxElectricTorque
                
                if ecvtDriveMode ~= 1 then
                    -- Consume available primary EV power first (from whatever is available from battery)
                    requestedElectricOutputTorque = min(torqueStillNeeded, availablePriorityEvOutputTorque) + creepTorque
                    -- Have to adjust for creepTorque here so as not to confuse the following calculations
                    torqueStillNeeded = torqueStillNeeded + creepTorque - requestedElectricOutputTorque
                end
                
                -- If we would use too much secondary EV power, start the ICE (or keep it on)
                local secondaryPowerDemand = mathutil.motorTorqueToKw(torqueStillNeeded, gearbox.motorGenerator2AV)
                    
                if secondaryPowerDemand > M.maxEvOnlySecondaryPower then
                    requestEngine("secondary power exceeded")
                end
                
                -- Accomplish any torque still needed with ICE if it's running
                if state == "running" then
                    requestedEngineOutputTorque = min(torqueStillNeeded, torqueData.maxEngineTorque)
                    targetEngineRpmCoef = M.engineRpmVsThrottle:get(normalizedThrottle * 100)
                    
                    -- While the engine is working on revving up, the main drive motor should compensate for it
                    -- so that the driver feels a more immediate acceleration response.
                    -- It's okay if we do this even if the battery is low (it's a short period of time it will be drawing power)
                    local immediateMG2CompensationNeededOnOutput = max(requestedEngineOutputTorque - actualEngineTorqueOnOutput, 0)
                    local immediateMG2CompensationNeeded = immediateMG2CompensationNeededOnOutput / gearbox.finalMG2GearRatio
                    
                    torqueStillNeeded = torqueStillNeeded - requestedEngineOutputTorque + immediateMG2CompensationNeeded
                end
                
                -- If torque is *still* needed, accomplish it with secondary EV torque (from ICE and battery reserve)
                requestedElectricOutputTorque = requestedElectricOutputTorque + min(torqueStillNeeded, availableSecondaryEvOutputTorque)
                torqueStillNeeded = torqueStillNeeded - min(torqueStillNeeded, availableSecondaryEvOutputTorque)
                
                -- If torque is ***still*** needed, start the ICE. We're probably in EV mode with a very low battery
                if gear ~= "R" and (torqueStillNeeded > 1e-20 or requestedElectricOutputTorque > torqueData.maxElectricTorque) then
                    requestEngine("electric torque exhausted")
                end
                
                requestedElectricOutputTorque = min(requestedElectricOutputTorque, torqueData.maxElectricTorque)
                
                if gear ~= "R" and not miscEngineDemand then
                    if throttle > M.maxEvOnlyThrottle then
                        requestEngine("ev throttle limit exceeded")
                    end
                    if brakeInput >= 0.25 and throttle > 0.02 then
                        requestEngine("launch control")
                    end
                end
                
                if brakeInput > 0.1 or electrics.values.parkingbrake > 0 then
                    requestedElectricOutputTorque = 0
                end
                
                debugutil.putValue("Available EV Torque (Priority)", fnum(availablePriorityEvOutputTorque))
                debugutil.putValue("Available EV Torque (Secondary)", fnum(availableSecondaryEvOutputTorque))
            else
                requestedTorque = 0
                requestedElectricOutputTorque = 0
                requestedEngineOutputTorque = maprange(normalizedThrottle, M.evCreepThrottle, 1, 0, M.maxParkEngineTorqueCoef * torqueData.maxEngineTorque)
                targetEngineRpmCoef = normalizedThrottle
                
                if throttle > 0.02 then
                    requestEngine("throttle in park")
                end
            end
            
            local targetMG1Torque = 0
            
            if state == "startup" then
                targetEngineRpm = minRpm
            elseif state == "running" then
                -- The torque of MG1 has to match the torque that the engine is currently producing
                -- It also has to be biased to allow the engine to reach its target RPM faster
                
                local requestedEngineTorqueCarrier = 0
                local engineTorqueForChargeTorque = 0
                local chargePowerCoef = 0
                
                if (speedMph < 0.5 and brakeInput > 0.1) or gear == "P" then
                    -- Charge while stopped. Charge power is achieved using pure MG1 drag on ICE.
                    
                    if isInChargeup then
                        requestedChargePower = M.chargeupMinPower
                        chargePowerCoef = 1.0
                    else
                        if gear == "P" then
                            chargePowerCoef = maprange(batteryLevel, M.minBatteryPowerLevel - 0.05, M.minBatteryPowerLevel, 1, 0.5)
                        else
                            chargePowerCoef = 0.25
                        end
                        
                        requestedChargePower = M.maxChargePowerParked * chargePowerCoef * maprange(batteryWorkingRatio, 0.5, 0)
                    end
                    
                    local requestedChargeTorque = mathutil.motorKwToTorque(requestedChargePower / M.regenEfficiency, gearbox.motorGenerator1AV)
                    
                    engineTorqueForChargeTorque = abs(gearbox:getCarrierTorqueFromSunTorque(requestedChargeTorque))
                    requestedEngineTorqueCarrier = requestedEngineTorqueCarrier + engineTorqueForChargeTorque
                    
                    debugutil.putValue("requestedChargeTorque",requestedChargeTorque)
                    debugutil.putValue("engineTorqueForChargeTorque",engineTorqueForChargeTorque)
                    debugutil.putValue("requestedEngineTorqueCarrier",requestedEngineTorqueCarrier)
                else
                    local chargePowerThrottleCoef = maprange(normalizedThrottle, 0.5, 0.8, 1, 0)
                    local chargePowerBatteryCoef = maprange(batteryWorkingRatio, M.targetRelativeBatteryLevel, M.targetRelativeBatteryLevel - M.targetBatteryLevelTolerance)
                    local chargeOutputTorque = 0
                    local maxChargePower = M.maxChargePowerVsSpeed:get(speedMph)
                    
                    chargePowerCoef = chargePowerThrottleCoef * chargePowerBatteryCoef
                    requestedChargePower = chargePowerCoef * maxChargePower
                    
                    if ecvtDriveMode == 1 then
                        -- Charge during mid-to-high-speed cruising. Charge power is achieved using regen drag on MG2, with
                        -- necessary accomodation with extra engine torque
                        
                        -- Positive value representing the amount of regen torque that will be needed to achieve requested charge power
                        chargeOutputTorque = mathutil.motorKwToTorque(requestedChargePower / M.regenEfficiency, gearbox.motorGenerator2AV) * gearbox.finalMG2GearRatio
                    else
                        -- Charge during lower speed cruising. Charge power is achieved by redistributing power from MG2 to MG1
                        
                        -- Extra engine load torque needed to achieve needed charge power
                        local chargeSunTorque = mathutil.motorKwToTorque(requestedChargePower / M.regenEfficiency, gearbox.motorGenerator1AV)
                        
                        -- Positive value representing the amount of regen torque that will be needed to achieve requested charge power
                        chargeOutputTorque = abs(gearbox:getRingTorqueFromSunTorque(chargeSunTorque)) * gearbox.finalRingGearRatio
                    end
                    
                    -- Reallocate as much torque as possible
                    local newRequestedEngineOutputTorque = min(torqueData.maxEngineTorque, requestedEngineOutputTorque + chargeOutputTorque)
                    local difference = max(0, newRequestedEngineOutputTorque - requestedEngineOutputTorque)
                    
                    requestedEngineOutputTorque = newRequestedEngineOutputTorque
                    requestedElectricOutputTorque = requestedElectricOutputTorque - difference
                end
                
                if isInDriveGear then
                    requestedEngineTorqueCarrier = requestedEngineTorqueCarrier + abs(gearbox:getCarrierTorqueFromRingTorque(requestedEngineOutputTorque / gearbox.finalRingGearRatio))
                end
                
                local achievableEngineTorqueSun = abs(gearbox:getSunTorqueFromCarrierTorque(achievableEngineTorque * gearbox.inputGearRatio))
                local requestedEngineTorqueSun = abs(gearbox:getSunTorqueFromCarrierTorque(requestedEngineTorqueCarrier))
                
                targetMG1Torque = requestedEngineTorqueSun
                
                local minTargetRpm = minRpm
                local maxMG1TorqueCarrierAV = gearbox:getCarrierAVFromSunAV(gearbox.mg1PeakEfficiencyAv)
                local maxMG1TorqueRpm = maxMG1TorqueCarrierAV * gearbox.inputGearRatio * constants.avToRPM
                local maxTargetRpm = min(maxRpm, max(minTargetRpm, maxMG1TorqueRpm))
                
                targetEngineRpm = maprange(targetEngineRpmCoef, 0, 1, minTargetRpm, maxTargetRpm)
                
                if targetEngineRpm > minTargetRpm + 50 then
                    requestEngine("rpm")
                end
                
                debugutil.putValue("Min Target RPM", minTargetRpm)
                debugutil.putValue("Max Target RPM", maxTargetRpm)
                debugutil.putValue("Requested Engine Torque (Sun, Carrier)", fnum(requestedEngineTorqueSun) .. ", " .. fnum(requestedEngineTorqueCarrier))
                debugutil.putValue("MG1 Max Torque Carrier AV", maxMG1TorqueCarrierAV)
                debugutil.putValue("MG1 Max Torque RPM", maxMG1TorqueRpm)
                debugutil.putValue("Achievable Engine Torque", fnum(achievableEngineTorque) .. " (" .. fnum(achievableEngineTorqueSun) .. ")")
                debugutil.putValue("Requested Charge Power", requestedChargePower .. " (" .. fnum(chargePowerCoef * 100) .. "%)")
            end
                
            if state == "running" then
                -- Calculate RPM load bias
                local rpmError = maprange(actualRpm, targetEngineRpm - M.engineLoadRpmTolerance, targetEngineRpm + M.engineLoadRpmTolerance, -1, 1)
                local loadBiasCoef = rpmError * (rpmError > 0 and M.engineMaxPositiveLoadBias or M.engineMaxNegativeLoadBias)
                local loadBiasTorque = loadBiasCoef * gearbox:getMaxMG1Torque(true)
                
                engineLoad = targetMG1Torque + loadBiasTorque
                
                -- Calculate engine deloading
                local availableEngineDeloadBattery = maprange(batteryWorkingRatio, 0, 0.1, 0, 1)
                local availableEngineDeloadRegen = maprange(requestedRegen, 0, 1, 0, 1)
                local availableEngineDeload = M.maxEngineDeload * mathutil.maxall(availableEngineDeloadBattery, availableEngineDeloadRegen)
                
                -- Apply engine deloading
                engineDeload = maprange(desiredRegen, 0, 0.05, 0, availableEngineDeload)
                engineLoad = engineLoad - engineDeload
                
                -- Limit engine load based on certain criteria
                local maxEngineLoadStall = maprange(electrics.values.rpm, engine.idleRPM - 200, engine.idleRPM)
                local maxEngineLoadOverspeed = maprange(mathutil.maxall(mg2OverspeedCoef, vehicleOverspeedCoef), 1, 0)
                local maxEngineLoadBattery = ecvtDriveMode == 1 and 1 or maprange(batteryLevel, 1, 0.98)
                
                engineLoad = engineLoad * mathutil.minall(maxEngineLoadStall, maxEngineLoadOverspeed, maxEngineLoadBattery)
                
                if highSpeedCoast and ecvtDriveMode == 0 then
                    -- Have to flip engine load sign in high speed coast mode
                    engineLoad = -engineLoad
                end
                
                -- Engine throttle is controlled by the desired output torque of the engine
                -- Engine RPM is controlled by the throttle pedal
                -- Actual load on the engine is biased to achieve the desired RPM
                local targetIceThrottle = M.driveThrottlePid:get(targetEngineRpm, actualRpm, dt)
                
                M.driveThrottlePid:setILock(false)
                iceThrottle = clamp(targetIceThrottle, 0, 1)
            
                debugutil.putValue("RPM Error", rpmError)
                debugutil.putValue("Load Bias", loadBiasTorque)
                debugutil.putValue("Target ICE Throttle", targetIceThrottle)
            else
                M.driveThrottlePid:reset()
                M.driveThrottlePid:setILock(true)
                M.engineThrottleSmoother:reset()
                
                -- To shut down engine faster, we apply engine load proportional to the engine RPM (but only if MG1 is spinning positive)
                local engineLoadShutdown = maprange(actualRpm, 10, engine.idleRPM, 0, M.engineSpooldownTorque)
                
                if gearbox.sunGearAV > gearbox.generatorMinFullLoadAV then
                    engineLoad = engineLoadShutdown
                else
                    engineLoad = 0
                end
                
                iceThrottle = 0
                engineDeload = 0
                targetEngineRpm = minRpm
            end
            
            -- Calculate actual torques
            local actualTorque = gearbox.outputTorque1
            local actualTorqueEV = gearbox.motorGenerator2Torque * gearbox.finalMG2GearRatio
            local actualTorqueICE = -gearbox:getRingTorqueFromCarrierTorque(actualEngineTorqueOnCarrier) * gearbox.finalRingGearRatio
            
            if state ~= "running" then
                ecvtDriveMode = 0
                M.latchDriveModeZero:reset()
                M.latchDriveModeOne:reset()
            end
            
            -- Determine actual requested torque from requested regen coef
            local regenTorque = requestedRegen * gearboxMaxMG2TorqueUnchecked * gearbox.finalMG2GearRatio
            
            -- Determine MG2 adjustment/compensation values
            if state == "startup" and isInDriveGear then
                local starterTorqueRingCompensation = gearbox:getRingTorqueFromSunTorque(gearbox.motorGenerator1Torque)
                local starterTorqueFinalCompensation = starterTorqueRingCompensation * gearbox.finalRingGearRatio
                local starterTorqueMG2Compensation = starterTorqueFinalCompensation / gearbox.finalMG2GearRatio
                
                mg2TorqueAdjustment = starterTorqueMG2Compensation
            end
            
            if state == "evmode" then
                -- Use both motors to achieve the requested torque
                
                local maxMG2TorqueToOutput = gearboxMaxMG2Torque * gearbox.finalMG2GearRatio
                local neededMG2TorqueToOutput = min(requestedElectricOutputTorque, maxMG2TorqueToOutput)
                local neededMG1CompensationToOutput = gear == "R" and 0 or max(0, requestedElectricOutputTorque - maxMG2TorqueToOutput)
                local neededMG2Torque = (abs(neededMG2TorqueToOutput) - regenTorque) / gearbox.finalMG2GearRatio
                local neededMG1Torque = abs(gearbox:getSunTorqueFromRingTorque(neededMG1CompensationToOutput / gearbox.finalRingGearRatio))
                
                mg2Torque = neededMG2Torque
                mg1Torque = neededMG1Torque
                mg1Dir = -1
            else
                local neededMG2Torque = (requestedElectricOutputTorque - regenTorque) / gearbox.finalMG2GearRatio
                
                if ecvtDriveMode == 0 then
                    -- engine load can be achieved using gen drag on MG1
                    -- charging is done by adding load to the engine
                    
                    mg1Torque = -engineLoad
                    mg1Dir = highSpeedCoast and -1 or 1
                    mg2Torque = neededMG2Torque + mg2TorqueAdjustment
                else
                    -- MG1 spinning too slow (or backwards) for gen drag; use negative power instead
                    -- in this case, we have to regen off of MG2 to provide power for MG1
                    -- (thus we have to limit MG1 to what MG2 is able to counteract)
                    
                    local engineLoadPower = abs(mathutil.motorTorqueToKw(max(0, engineLoad), gearbox.motorGenerator1AV))
                    local batteryPowerAvailableForEngineLoad = availablePriorityEvPower - max(0, gearbox.motorGenerator2Power) -- can use priority power to apply engine load
                    local neededPowerForEngineLoad = max(0, engineLoadPower - batteryPowerAvailableForEngineLoad)
                    local neededRegenTorqueForEngineLoad = abs(mathutil.motorKwToTorque(neededPowerForEngineLoad, gearbox.motorGenerator2AV))
                    
                    mg1Torque = abs(engineLoad)
                    mg1Dir = -sign(engineLoad)
                    mg2Torque = neededMG2Torque - neededRegenTorqueForEngineLoad + mg2TorqueAdjustment
                    
                    requestEngine("eCVT mode")
                end
            end
            
            if state == "evmode" then
                ecvtDriveMode = 0
            else
                -- Determine which drive mode we should be in next based on whether or not MG1 can provide generation drag
                local modeZeroLatchTest = requestedEngineOutputTorque < 1e-3 or normalizedThrottle < 1e-3 or gearbox.sunGearAV > gearbox.generatorMinFullLoadAV + 150
                local modeOneLatchTest = requestedEngineOutputTorque >= 1e-3 and normalizedThrottle >= 1e-3 and gearbox.sunGearAV < gearbox.generatorMinFullLoadAV * 1.5
                
                if M.latchDriveModeOne:test(modeOneLatchTest, dt) then
                    if ecvtDriveMode == 0 then
                        M.latchDriveModeZero:reset()
                        ecvtDriveMode = 1
                    end
                end
                if M.latchDriveModeZero:test(modeZeroLatchTest, dt) then
                    if ecvtDriveMode == 1 then
                        M.latchDriveModeOne:reset()
                        ecvtDriveMode = 0
                    end
                end
                
                -- "High speed coast" mode forces eCVT into mode 0 even when it should be in mode 1 so that a light re-application
                -- of the throttle will go to EV first before jumping back to mode 1 ICE operation
                highSpeedCoast = gearbox.sunGearAV < gearbox.generatorMinFullLoadAV * 1.5 and modeZeroLatchTest
            end
            
            debugutil.putValue("Requested Torque", fnum(requestedTorque) .. " (" .. fnum(requestedElectricOutputTorque) .. " / " .. fnum(requestedEngineOutputTorque) .. ")")
            debugutil.putValue("Actual Torque", fnum(actualTorque) .. " (" .. fnum(actualTorqueEV) .. " / " .. fnum(actualTorqueICE) .. ")")
            debugutil.putValue("Available MG1 Power", fnum(availableMG1Power))
            debugutil.putValue("Available MG2 Power", fnum(availableMG2Power))
            debugutil.putValue("ICE Throttle", iceThrottle)
            debugutil.putValue("Requested Regen", requestedRegen)
            debugutil.putValue("Engine Load", engineLoad)
            debugutil.putValue("Engine Deload", engineDeload)
            debugutil.putValue("Target MG1 Torque", targetMG1Torque)
            debugutil.putValue("Min RPM", minRpm)
            debugutil.putValue("Max RPM", maxRpm)
            debugutil.putValue("Target RPM", targetEngineRpm)
            debugutil.putValue("Actual RPM", actualRpm)
            debugutil.putValue("Actual ICE Throttle", electrics.values.iceThrottle)
            debugutil.putValue("Actual Engine Torque", fnum(actualEngineTorque) .. " (" .. fnum(actualEngineTorqueOnSun) .. " / " .. fnum(actualEngineTorqueOnRing) .. ")")
            debugutil.putValue("eCVT Drive Mode", ecvtDriveMode)
            debugutil.putValue("High Speed Coast", highSpeedCoast)
        end
        
        debugutil.putValue("Input Clutch", fnum(electrics.values[clutch.electricsClutchRatioName]))
        
        gearbox:setMG1TorqueDirection(mg1Dir)
        gearbox:setRequestedMG1Torque(mg1Torque)
        gearbox:setRequestedMG2Torque(mg2Torque)
        
        -- Calculate motor power restrictions
        local mg1GeneratedPower = max(-gearbox.motorGenerator1Power, 0)
        local mg2GeneratedPower = max(-gearbox.motorGenerator2Power, 0)
        
        availableMG1Power = min(gearbox.mg1PowerRating, availableBatteryPower + mg2GeneratedPower)
        
        if ecvtDriveMode == 0 then
            availableMG2Power = min(gearbox.mg2PowerRating, availableBatteryPower + mg1GeneratedPower)
        else
            availableMG2Power = min(gearbox.mg2PowerRating, availableBatteryPower)
        end
        
        -- Limit the amount of power MG1 and MG2 are allowed to consume (so they don't overdraw the battery)
        gearbox:setMG1PowerOverride(M.mg1MaxPowerSmoother:get(availableMG1Power, dt))
        gearbox:setMG2PowerOverride(M.mg2MaxPowerSmoother:get(availableMG2Power, dt))
        
        updateState(dt)
        updateBattery(dt)
    end
    
    electrics.values.iceThrottle = M.engineThrottleSmoother:get(iceThrottle, dt)

    if battery then
        debugutil.putValue("Battery SOC %", tostring(battery.remainingRatio * 100))
    end

    debugutil.putValue("Input Throttle", (throttle or 0))
    debugutil.putValue("ICE Throttle (True)", electrics.values.iceThrottle)
    debugutil.putValue("Throttle Aggr.", throttleAggressiveness)

    if nextState then
        state = nextState
    end

    debugutil.putValue("ICE Torque", engine.outputTorque1)
    debugutil.putValue("ICE HP", mathutil.motorTorqueToKw(engine.outputTorque1, engine.outputAV1) * 1.34102)
    debugutil.putValue("Charge Throttle", chargeThrottle)
    debugutil.putValue("MG2 Adjustment", mg2TorqueAdjustment)

    updateGUI(dt)
    updateSounds(dt)
end

local function init(jbeamData)
    M.jbeamData = jbeamData
    
    M.startEvMode = jbeamData.startEvMode == true
    M.forceEvMode = jbeamData.forceEvMode == true
    
    M.maxEngineOnPriorityRatio = jbeamData.maxEngineOnPriorityRatio or 0.5
    M.maxEvOnlyPriorityPower = jbeamData.maxEvOnlyPriorityPower or 30 -- kW
    M.maxEvOnlySecondaryPower = jbeamData.maxEvOnlySecondaryPower or 25 -- kW
    M.maxEvOnlySpeed = jbeamData.maxEvOnlySpeed or 65 -- MPH
    M.maxEvOnlyThrottle = jbeamData.maxEvOnlyThrottle or 0.65 -- ratio
    M.maxEvOnlyThrottleAggressiveness = jbeamData.maxEvOnlyThrottleAggressiveness or 20
    M.maxEvOnlyPitch = jbeamData.maxEvOnlyPitch or math.pi -- 180 degrees pitch by default; will never happen (effectively disabled)
    M.evOnlyThrottleDetent = jbeamData.evOnlyThrottleDetent
    M.engineShutOffThrottleThreshold = jbeamData.engineShutOffThrottleThreshold or 0.10
    
    M.minRegenFadeSpeed = jbeamData.minRegenFadeSpeed or 8 -- MPH
    M.maxRegenFadeSpeed = jbeamData.maxRegenFadeSpeed or 38 -- MPH
    
    M.idlePowerDraw = jbeamData.idlePowerDraw or 1 -- kW
    M.minBatteryPowerLevel = 0.25 -- ratio; low end of max-kW slope
    M.maxBatteryPowerLevel = 0.85 -- ratio; high end of max-kW slope
    M.targetRelativeBatteryLevel = jbeamData.targetRelativeBatteryLevel or 0.5 -- 50%
    M.targetBatteryLevelTolerance = jbeamData.targetBatteryLevelTolerance or 0.15 -- 15%
    M.maxFastChargePower = jbeamData.maxFastChargePower or 30 -- kW
    M.fastChargeRpm = jbeamData.fastChargeRpm
    
    M.enableAutoStop = jbeamData.enableAutoStop ~= false
    M.autoStopBrakeThreshold = jbeamData.autoStopBrakeThreshold or 0
    M.autoStopBrakeTime = jbeamData.autoStopBrakeTime or 0
    M.engineMinRunTime = jbeamData.engineMinRunTime or 4 -- seconds
    
    M.engineMaxRpm = jbeamData.engineMaxRpm
    M.engineStartupIdleTime = jbeamData.engineStartupIdleTime or 0.1
    M.engineStarterTorque = jbeamData.engineStarterTorque
    M.engineSpooldownTorque = jbeamData.engineSpooldownTorque or 50
    M.maxChargePowerParked = jbeamData.maxChargePowerParked or 10
    M.chargeupMinPower = jbeamData.chargeupMinPower or 10
    M.engineMinIdleRpm = jbeamData.engineMinIdleRpm
    M.engineMinChargeRpm = jbeamData.engineMinChargeRpm
    M.engineLoadRpmTolerance = jbeamData.engineLoadRpmTolerance or 350
    M.engineMaxPositiveLoadBias = jbeamData.engineMaxPositiveLoadBias or 0.85
    M.engineMaxNegativeLoadBias = jbeamData.engineMaxNegativeLoadBias or 0.85
    M.engineTorqueToRpmExponent = jbeamData.engineTorqueToRpmExponent or 5.5
    M.maxEngineDeload = jbeamData.maxEngineDeload or 0.01
    M.engineDeloadMaxIgnitionCut = jbeamData.engineDeloadMaxIgnitionCut or 0.4
    M.maxParkEngineTorqueCoef = jbeamData.maxParkEngineTorqueCoef or 0.5
    M.maxVehicleSpeed = jbeamData.maxVehicleSpeed
    M.maxVehicleSpeedReverse = jbeamData.maxVehicleSpeedReverse or 20
    M.maxVehicleSpeedFalloff = jbeamData.maxVehicleSpeedFalloff or 5
    M.maxVehicleSpeedFalloffReverse = jbeamData.maxVehicleSpeedFalloffReverse or M.maxVehicleSpeedFalloff
    
    M.minThrottleAggressivenessRamp = jbeamData.minThrottleAggressivenessRamp or 1.5
    M.maxThrottleAggressivenessRamp = jbeamData.maxThrottleAggressivenessRamp or 8
    M.minThrottleAggressivenessFactor = jbeamData.minThrottleAggressivenessFactor or 1
    M.maxThrottleAggressivenessFactor = jbeamData.maxThrottleAggressivenessFactor or 10
    M.throttleCurveExponent = jbeamData.throttleCurveExponent or 1.1
    
    M.maxRegenCoef = jbeamData.maxRegenCoef or 0.65
    M.regenEfficiency = jbeamData.regenEfficiency or 0.93
    M.regenPercentOfBrakingForce = jbeamData.regenPercentOfBrakingForce or 0.2
    
    M.evCreepThrottle = jbeamData.evCreepThrottle or 0.25
    M.evCoastRegen = jbeamData.evCoastRegen or 0.05
    M.hillAssistBrakeHoldAmount = jbeamData.hillAssistBrakeHoldAmount or 0.15 -- %
    M.minHillAssistPitch = jbeamData.minHillAssistPitch or 0.0523599 -- rads; 3 degrees
    M.maxHillAssistPitch = jbeamData.maxHillAssistPitch or 0.20944 -- rads; 12 degrees
    
    M.inverterMaxVolumeKw = jbeamData.inverterMaxVolumeKw or 50
    M.inverterBaseVolume = jbeamData.inverterBaseVolume or 0.65
    M.inverterPitchMinSpeed = jbeamData.inverterPitchMinSpeed or 1
    M.inverterPitchMaxSpeed = jbeamData.inverterPitchMaxSpeed or 1100
    M.inverterMinPitch = jbeamData.inverterMinPitch or 0.01
    M.inverterMaxPitch = jbeamData.inverterMaxPitch or 2.0
    M.motorBaseVolume = jbeamData.motorBaseVolume or 0.10
    M.motorVolumeMinSpeed = jbeamData.motorVolumeMinSpeed or 0.2
    M.motorVolumeMaxSpeed = jbeamData.motorVolumeMaxSpeed or 50
    M.motorPitchMinSpeed = jbeamData.motorPitchMinSpeed or 0
    M.motorPitchMaxSpeed = jbeamData.motorPitchMaxSpeed or 95
    M.motorMinPitch = jbeamData.motorMinPitch or 0.3
    M.motorMaxPitch = jbeamData.motorMaxPitch or 3.4
    
    -- Charge power vs speed curve
    M.maxChargePowerVsSpeed = simpleCurve.createSimpleCurve(jbeamData.maxChargePowerVsSpeed, "speed", "maxChargePower", {
        {"speed", "maxChargePower"},
        {0, 10},
        {45, 10},
        {65, 25},
    })
    
    -- Throttle vs RPM curve
    M.engineRpmVsThrottle = simpleCurve.createSimpleCurve(jbeamData.engineRpmVsThrottle, "throttle", "rpmCoef", {
        {"throttle", "rpmCoef"},
        {0, 0},
        {75, 1},
    })
    
    M.driveThrottlePid = pid.newPid(
        jbeamData.throttlePidPFactor or 0.4,
        jbeamData.throttlePidIFactor or 0.95,
        jbeamData.throttlePidDFactor or 0.11
    ):setScale(1 / (jbeamData.throttlePidRPMScale or 1000)):clampOutput(0, 1):setDynamicILock(0.6, false):setName("throttle")--:enableDebug()
    M.chargeThrottlePid = pid.newPid(
        jbeamData.chargeThrottlePidPFactor or 2.5, 
        jbeamData.chargeThrottlePidIFactor or 5.5
    ):setScale(1 / 1000):clampOutput(0, 1)

    -- Lots of temporal smoothers!
    M.throttleSmoother = newTemporalSmoothingNonLinear(
        jbeamData.throttleSmoothingIn or 1.5,
        jbeamData.throttleSmoothingOut or 1.5)
    M.throttleLimiterSmoother = newTemporalSmoothingNonLinear(jbeamData.throttleLimitSmoothing or 0.95)
    M.engineThrottleSmoother = newTemporalSmoothingNonLinear(jbeamData.engineThrottleSmoothing or 8)
    M.engineActualRpmSmoother = newTemporalSmoothingNonLinear(
        jbeamData.engineActualRpmSmoothingIn or 12.5,
        jbeamData.engineActualRpmSmoothingOut or 12.5
    )
    M.engineStarterSmoother = newTemporalSmoothingNonLinear(
        jbeamData.engineStarterSmoothingIn or 7,
        jbeamData.engineStarterSmoothingOut or 4)
    M.mg1MaxPowerSmoother = newTemporalSmoothingNonLinear(jbeamData.motorPowerSmoothing or 0.2)
    M.mg2MaxPowerSmoother = newTemporalSmoothingNonLinear(jbeamData.motorPowerSmoothing or 0.2)
    M.mg1PowerGeneratedSmoother = newTemporalSmoothingNonLinear(jbeamData.motorGeneratedPowerSmoothing or 0.5)
        
    -- Sound smoothers
    M.mg1PitchSmoother = newTemporalSmoothingNonLinear(6)
    M.mg2PitchSmoother = newTemporalSmoothingNonLinear(6)
    
    -- Latches
    M.latchDriveModeZero = latch.newTimedLatch(jbeamData.modeZeroLatchTime or 0.25)
    M.latchDriveModeOne  = latch.newTimedLatch(jbeamData.modeOneLatchTime or 0.25, true)

    -- Pre-fill debug table with empty strings to set up proper order
    debugutil.clearValues()
    debugutil.putValue("Coolant Temp.", "")
    debugutil.putValue("Engine demanded on", "")
    debugutil.putValue("Engine on reason", "")
    debugutil.putValue("Engine triggered on by", "")
    debugutil.putValue("State", "")
    debugutil.putValue("Battery power (kW)", "")
    debugutil.putValue("Generator power (kW)", "")
    debugutil.putValue("Battery SOC %", "")
    debugutil.putValue("Battery SOC % (usable)", "")
    debugutil.putValue("Max Regen", "")
    debugutil.putValue("Friction % of full brakes", "")
    debugutil.putValue("Regen % of full brakes", "")
    debugutil.putValue("Regen % of full regen", "")
    debugutil.putValue("Requested Regen", "")
    debugutil.putValue("Charge Throttle", "")
    debugutil.putValue("Input Throttle", "")
    debugutil.putValue("Throttle Aggr.", "")
    debugutil.putValue("ICE Throttle", "")
    debugutil.putValue("Throttle Rate", "")
    debugutil.putValue("Throttle % EV thresh.", "")
    debugutil.putValue("Requested Torque", "")
    debugutil.putValue("Actual Torque", "")
    debugutil.putValue("Achievable Engine Torque", "")
    debugutil.putValue("Actual Engine Torque", "")
    debugutil.putValue("Available EV Torque (Priority)", "")
    debugutil.putValue("Available EV Torque (Secondary)", "")
    debugutil.putValue("RPM Error", "")
    debugutil.putValue("Load Bias", "")
    debugutil.putValue("Target MG1 Torque", "")
    debugutil.putValue("Engine Load", "")
    debugutil.putValue("Min RPM", "")
    debugutil.putValue("Max RPM", "")
    debugutil.putValue("Min Target RPM", "")
    debugutil.putValue("Max Target RPM", "")
    debugutil.putValue("Target RPM", "")
    debugutil.putValue("Actual RPM", "")
    debugutil.putValue("Target ICE Throttle", "")
    debugutil.putValue("Actual ICE Throttle", "")
    debugutil.putValue("ICE Throttle (True)", "")
    debugutil.putValue("ICE Torque", "")
    debugutil.putValue("ICE HP", "")
    debugutil.putValue("MG2 Adjustment", "")
end

local function reset()
    timeSinceInit = 0
    state = "init"
    nextState = nil
    
    M.chargeThrottlePid:reset()
    M.driveThrottlePid:reset()
    M.latchDriveModeZero:reset()
    M.latchDriveModeOne:reset()

    init(M.jbeamData)
    M.initSecondStage()
end

local function initSounds()
    sounds.mg1Whine =
        obj:createSFXSource(
        "vehicles/hybrid_common/sounds/inverter.ogg",
        "AudioDefaultLoop3D",
        "mg1Whine",
        engine.engineNodeID
    )
    
    sounds.mg2Whine =
        obj:createSFXSource(
        "vehicles/hybrid_common/sounds/inverter.ogg",
        "AudioDefaultLoop3D",
        "mg2Whine",
        engine.engineNodeID
    )

    sounds.motorRamp =
        obj:createSFXSource(
        "vehicles/hybrid_common/sounds/motor_ramp.ogg",
        "AudioDefaultLoop3D",
        "motorRamp",
        engine.engineNodeID
    )
end

local function initSecondStage()
    engine = powertrain.getDevice("mainEngine")
    gearbox = powertrain.getDevice("gearbox")
    clutch = powertrain.getDevice("clutch")
    battery = energyStorage.getStorage("tractionBattery")

    if not gearbox or gearbox.type ~= "ecvtGearbox" then
        gearbox = nil
    end

    if battery then
        M.minBatteryPowerLevel = battery.jbeamData.minWorkingSoc or M.minBatteryPowerLevel
        M.maxBatteryPowerLevel = battery.jbeamData.maxWorkingSoc or M.maxBatteryPowerLevel

        if not M.maxFastChargePower then
            M.maxFastChargePower = battery.jbeamData.maxPowerInput
        end
    end
    
    if engine then
        engineTorqueData = engine:getTorqueData()
        
        if not M.jbeamData.engineMaxRpm then
            M.engineMaxRpm = engineTorqueData.maxPowerRPM
        end
        
        if not M.engineMinIdleRpm then
            M.engineMinIdleRpm = engine.idleRPM + 200
        end
    end
    
    if engine and gearbox then
        calculateTorqueStatistics()
        printTorqueData()
        
        if not M.engineStarterTorque then
            M.engineStarterTorque = gearbox.mg1TorqueRating
        end
    end

    for i = 0, wheels.wheelCount - 1, 1 do
        if not initialWheelBrakeSplit[i] then
            initialWheelBrakeSplit[i] = {wheels.wheels[i].brakeInputSplit, wheels.wheels[i].brakeSplitCoef}
        end
    end

    if M.startEvMode then
        electrics.values.evMode = 1
    end
end

M.updateGFX = updateGFX
M.init = init
M.initSounds = initSounds
M.initSecondStage = initSecondStage
M.reset = reset
M.getTorqueData = getTorqueData
M.printTorqueData = printTorqueData

return M
